#include <intrin.h>
#include "driver.h"


#define AlignAddress(p, a) (ULONG64)(((ULONG64)(p)+((ULONG64)a - 1)) & ~((ULONG64)a - 1))

//#define USE_POOL

ULONG_PTR TrampolineInDriver = (ULONG_PTR)_syscall64 + 0x1000; // 0x100
ULONG_PTR SsdtTrampolineInDriver = (ULONG_PTR)_syscall64 + 0x1100; // 0x1200
ULONG_PTR ShadowSsdtTrampolineInDriver = (ULONG_PTR)_syscall64 + 0x2300; // 0x1E00
ULONG_PTR CallR10InDriver, CallR10InMdl, CallR10InSystem;
PULONG_PTR CallR10;

ULONG_PTR TrampolineInMdl;
ULONG_PTR SsdtTrampolineInMdl;
ULONG_PTR ShadowSsdtTrampolineInMdl;

ULONG64 PtrAddress[0x100];
PVOID Syscall64Pointer;


LONG SsdtTable[0x200];
ULONG64 SsdtAddress[0x200];

LONG ShadowSsdtTable[0x500];
ULONG64 ShadowSsdtAddress[0x500];

SERVICE_DESCRIPTOR_TABLE SsdtInDriver;
ULONG64 Nop[4];
SERVICE_DESCRIPTOR_TABLE_SHADOW ShadowSsdtInDriver;



PMDL AllocateMdl(IN PVOID From, IN SIZE_T Size, OUT PULONG_PTR To)
{
	NTSTATUS status = STATUS_SUCCESS;

	PVOID NewAddress;
	PMDL Mdl = IoAllocateMdl(From, (ULONG)Size, FALSE, TRUE, NULL);
	if (!Mdl)
		return NULL;
	PAGED_CODE();

	__try {
		MmProbeAndLockPages(Mdl, KernelMode, IoWriteAccess);
	}
	__except (EXCEPTION_EXECUTE_HANDLER) {
		IoFreeMdl(Mdl);
		return NULL;
	}

	if (Mdl->MdlFlags & (MDL_MAPPED_TO_SYSTEM_VA | MDL_SOURCE_IS_NONPAGED_POOL))
		NewAddress = Mdl->MappedSystemVa;
	else
		NewAddress = MmMapLockedPagesSpecifyCache(Mdl, KernelMode, MmCached, NULL, FALSE, NormalPagePriority);

	if (!NewAddress)
	{
		MmUnlockPages(Mdl);
		IoFreeMdl(Mdl);
		return NULL;
	}

	status = MmProtectMdlSystemAddress(Mdl, PAGE_EXECUTE_READWRITE);

	if (!NT_SUCCESS(status))
	{
		MmUnlockPages(Mdl);
		IoFreeMdl(Mdl);
		return NULL;
	}

	if (To)
		*To = (ULONG_PTR)NewAddress;

	return Mdl;
}

NTSTATUS InitializeFakeSyscall()
{
	NTSTATUS status = STATUS_SUCCESS;
	ULONG_PTR syscall64start = __readmsr(0xC0000082);
	ULONG_PTR syscall64end = 0;
	SIZE_T syscall64length = 0;

	if (!syscall64start)
		return STATUS_NOT_FOUND;

	// 0x1000 may be enough?
	for (SHORT i = 0;i < 0x1000;i++)
	{
		if (*(PUCHAR)(syscall64start + i) == 0xC3)
		{
			// end suffix
			if (*(PUCHAR)(syscall64start + i + 1) == 0x66 &&
				*(PUCHAR)(syscall64start + i + 2) == 0x66 &&
				*(PUCHAR)(syscall64start + i + 3) == 0x66 &&
				*(PUCHAR)(syscall64start + i + 4) == 0x66)
			{
				syscall64end = syscall64start + i;
				syscall64length = syscall64end - syscall64start + 1;
				break;
			}
		}
	}

	if (!syscall64end)
		return STATUS_NOT_FOUND;


	SIZE_T SyscallLength = 0x5100;
	// 		Trampoline = 0x100,
	// 		SsdtTrampoline = 0x1200,
	// 		ShadowSsdtTrampoline = 0x1E00;
	DbgPrint("_syscall64 length = %x\n", SyscallLength);

	// copy codes to our space.
	ULONG_PTR MdlVa;
	PMDL SyscallMdl = AllocateMdl((PVOID)(ULONG_PTR)&_syscall64, SyscallLength, &MdlVa);
	if (!SyscallMdl)
		return STATUS_UNSUCCESSFUL;

	// still many bugs
	RtlCopyMemory((PVOID)MdlVa, (PULONG_PTR)syscall64start, syscall64length);

	TrampolineInMdl = (ULONG_PTR)MdlVa + 0x1000;
	SsdtTrampolineInMdl = (ULONG_PTR)MdlVa + 0x1100;
	ShadowSsdtTrampolineInMdl = (ULONG_PTR)MdlVa + 0x2300;

	// fix rel addr in our syscall.
	FixSyscall64RelativeAddress(MdlVa,
		(ULONG_PTR)syscall64start,
		syscall64length);

	__debugbreak();

	MmUnmapLockedPages((PVOID)MdlVa, SyscallMdl);
	MmUnlockPages(SyscallMdl);
	IoFreeMdl(SyscallMdl);

	return status;
}


VOID FixOpcodeCmpRsiInDriver(ULONG_PTR AddressInDriver,
	ULONG_PTR AddressInSystem,
	ULONG_PTR AddressInMdl,
	ULONG64 Offset,
	PULONG_PTR Ptr)
{
	UNREFERENCED_PARAMETER(AddressInDriver);
	UNREFERENCED_PARAMETER(AddressInSystem);

	ULONG_PTR MdlOffset = AddressInMdl + Offset;

	DbgPrint("From %p found cmp rsi, qword ptr [nt!MmUserProbeAddress]\n", MdlOffset);

	// push rax
	*(PULONG_PTR)TrampolineInMdl = 0x50;
	//mov rax, qword ptr
	*(PULONG_PTR)(TrampolineInMdl + 1) = 0x058B48;

	// rel32
	*(PULONG_PTR)*Ptr = (ULONG64)&MmUserProbeAddress;
	ULONG32 Rel32 = (ULONG32)(*Ptr - (TrampolineInDriver + 8));
	*(PULONG32)(TrampolineInMdl + 4) = Rel32;

	// cmp rsi,qword ptr [rax]
	*(PULONG_PTR)(TrampolineInMdl + 8) = 0x303B48;
	// pop rax
	*(PUCHAR)(TrampolineInMdl + 11) = 0x58;
	// ret
	*(PUCHAR)(TrampolineInMdl + 12) = 0xC3;

	// call
	*(PUCHAR)MdlOffset = 0xE8;


	ULONG_PTR DriverOffset = AddressInDriver + Offset;
	Rel32 = (ULONG32)(TrampolineInDriver - (DriverOffset + 5));

	RtlCopyMemory((PULONG_PTR)(MdlOffset + 1), &Rel32, 4);

	// nop 2 bytes
	memset((PULONG_PTR)(MdlOffset + 5), 0x90, 2);

	TrampolineInDriver += 13;
	TrampolineInMdl += 13;
	*Ptr += 8;
}

VOID FixOpcodeCmovaeRsiInDriver(ULONG_PTR AddressInDriver,
	ULONG_PTR AddressInSystem,
	ULONG_PTR AddressInMdl,
	ULONG64 Offset,
	PULONG_PTR Ptr)
{
	UNREFERENCED_PARAMETER(AddressInDriver);
	UNREFERENCED_PARAMETER(AddressInSystem);

	ULONG_PTR MdlOffset = AddressInMdl + Offset;

	DbgPrint("From %p found cmovae rsi,qword ptr [nt!MmUserProbeAddress]\n", MdlOffset);

	// push rax
	*(PULONG_PTR)TrampolineInMdl = 0x50;
	//mov rax, qword ptr
	*(PULONG_PTR)(TrampolineInMdl + 1) = 0x058B48;

	// rel32
	*(PULONG_PTR)*Ptr = (ULONG64)&MmUserProbeAddress;
	ULONG32 Rel32 = (ULONG32)(*Ptr - (TrampolineInDriver + 8));
	*(PULONG32)(TrampolineInMdl + 4) = Rel32;

	// cmovae  rsi,qword ptr [rax]
	*(PULONG_PTR)(TrampolineInMdl + 8) = 0x30430F48;
	// pop rax
	*(PUCHAR)(TrampolineInMdl + 12) = 0x58;
	// ret
	*(PUCHAR)(TrampolineInMdl + 13) = 0xC3;

	// call
	*(PUCHAR)MdlOffset = 0xE8;

	ULONG_PTR DriverOffset = AddressInDriver + Offset;
	Rel32 = (ULONG32)(TrampolineInDriver - (DriverOffset + 5));

	RtlCopyMemory((PULONG_PTR)(MdlOffset + 1), &Rel32, 4);

	// nop 3 bytes
	memset((PULONG_PTR)(MdlOffset + 5), 0x90, 3);

	TrampolineInDriver += 14;
	TrampolineInMdl += 14;
	*Ptr += 8;
}

VOID FixOpcodeTestInDriver(ULONG_PTR AddressInDriver,
	ULONG_PTR AddressInSystem,
	ULONG_PTR AddressInMdl,
	ULONG64 Offset,
	PULONG_PTR Ptr)
{
	UNREFERENCED_PARAMETER(AddressInDriver);

	ULONG_PTR MdlOffset = AddressInMdl + Offset;

	DbgPrint("From %p found test dword ptr [nt!PerfGlobalGroupMask],40h\n", MdlOffset);

	ULONG32 Rel32 = *(PULONG32)(MdlOffset + 2);

	// push rax
	*(PULONG_PTR)TrampolineInMdl = 0x50;
	// mov rax, qword ptr
	*(PULONG_PTR)(TrampolineInMdl + 1) = 0x058B48;

	// rel32
	*(PULONG_PTR)*Ptr = AddressInSystem + Offset + 10 + Rel32;
	Rel32 = (ULONG32)(*Ptr - (TrampolineInDriver + 8));
	*(PULONG32)(TrampolineInMdl + 4) = Rel32;

	// test qword ptr [rax],40h
	*(PULONG_PTR)(TrampolineInMdl + 8) = 0x4000F748;
	*(PULONG_PTR)(TrampolineInMdl + 12) = '\x00\x00\x00';
	// pop rax
	*(PUCHAR)(TrampolineInMdl + 15) = 0x58;
	// ret
	*(PUCHAR)(TrampolineInMdl + 16) = 0xC3;

	// call
	*(PUCHAR)MdlOffset = 0xE8;

	ULONG_PTR DriverOffset = AddressInDriver + Offset;
	Rel32 = (ULONG32)(TrampolineInDriver - (DriverOffset + 5));

	RtlCopyMemory((PULONG_PTR)(MdlOffset + 1), &Rel32, 4);

	// nop 5 bytes
	memset((PULONG_PTR)(MdlOffset + 5), 0x90, 5);

	TrampolineInMdl += 17;
	TrampolineInDriver += 17;
	*Ptr += 8;
}

VOID FixOpcodeCallInDriver(ULONG_PTR AddressInDriver,
	ULONG_PTR AddressInSystem,
	ULONG_PTR AddressInMdl,
	ULONG64 Offset,
	PULONG_PTR Ptr)
{
	UNREFERENCED_PARAMETER(AddressInDriver);

	ULONG_PTR MdlOffset = AddressInMdl + Offset;

	DbgPrint("0xE8 call found at %p\n", MdlOffset);

	ULONG32 Rel32 = *(PULONG32)(MdlOffset + 1);

	// rel32+next
	ULONG_PTR TargetAddress;
	if (Rel32 > 0xffff0000)
		TargetAddress = (Rel32 ^ 0xffffffff00000000) + AddressInSystem + Offset + 5;
	else
		TargetAddress = Rel32 + AddressInSystem + Offset + 5;

	if (MmIsAddressValid((PVOID)TargetAddress))
	{
		// jmp qword ptr
		*(PUCHAR)TrampolineInMdl = 0xFF;
		*(PUCHAR)(TrampolineInMdl + 1) = 0x25;

		// target - opcode rel32
		*(PULONG_PTR)*Ptr = TargetAddress;
		Rel32 = (ULONG32)(*Ptr - (TrampolineInDriver + 6));
		*(PULONG32)(TrampolineInMdl + 2) = Rel32;

		// call rel32
		ULONG_PTR DriverOffset = AddressInDriver + Offset;
		Rel32 = (ULONG32)(TrampolineInDriver - (DriverOffset + 5));

		*(PULONG32)(MdlOffset + 1) = Rel32;

		TrampolineInMdl += 6;
		TrampolineInDriver += 6;
		*Ptr += 8;
	}
	else
	{
		DbgPrint("From %p MmIsAddressValid() failed on %p\n", MdlOffset, TargetAddress);
	}
}

NTSTATUS FixSyscall64RelativeAddress(ULONG_PTR MdlVa, ULONG_PTR SyscallOrigin, SIZE_T Length)
{
	NTSTATUS status = STATUS_SUCCESS;
	PUCHAR syscall = (PUCHAR)MdlVa;

	ULONG_PTR Ptr = (ULONG64)&PtrAddress; // size = 0x71

										  // lea r10,[nt!KeServiceDescriptorTable]
										  // lea r11,[nt!KeServiceDescriptorTableShadow]
										  // lea rdi, [nt!KeServiceDescriptorTableShadow]
	ULONG64 leaR10Offset = 0, leaR11Offset = 0, leaRdiOffset = 0;

	// KeServiceDescriptorTable 0x20
	// KeServiceDescriptorTableShadow 0x40

	// SSDT
	// 0x1b9 * (6 bytes jmp + 8 bytes address) = 0x181e
	// ~0x200 = 0x1c00
	// total = 0x1c20

	// Shadow SSDT
	// 0x46a * (6 bytes jmp + 8 bytes address) = 0x3dcc
	// ~0x500 = 0x4600
	// total = 0x4640

	/*

	[Opcode]------------------------[Offset]
	[Offset]---------[Mdl]

	*/



	typedef VOID(*_handler)(ULONG_PTR, ULONG_PTR, ULONG_PTR, ULONG64, /*PULONG_PTR, PULONG_PTR,*/ PULONG_PTR);

	_handler Handler = NULL;

	for (ULONG Offset = 0;Offset < Length;Offset++)
	{
		// Fix opcode call
		// Change it to jmp ptr xxxxxxxx
		if (syscall[Offset] == 0xE8)
		{
			//			__debugbreak();
			Handler = FixOpcodeCallInDriver;
		}

		// Fix opcode cmp rsi, qword ptr [nt!MmUserProbeAddress]
		// Change it to a call xxxxxxxx
		else if (syscall[Offset] == 0x48 &&
			syscall[Offset + 1] == 0x3B &&
			syscall[Offset + 2] == 0x35)
		{
			//			__debugbreak();
			Handler = FixOpcodeCmpRsiInDriver;
		}

		// Fix opcode cmovae rsi,qword ptr [nt!MmUserProbeAddress]
		// Change it to call xxxxxxxx
		else if (syscall[Offset] == 0x48 &&
			syscall[Offset + 1] == 0x0f &&
			syscall[Offset + 2] == 0x43 &&
			syscall[Offset + 3] == 0x35)
		{
			//			__debugbreak();
			Handler = FixOpcodeCmovaeRsiInDriver;
		}

		// Fix opcode test dword ptr [nt!PerfGlobalGroupMask],40h
		// Change it to a call xxxxxxxx
		else if (syscall[Offset] == 0xF7 &&
			syscall[Offset + 1] == 0x05 &&
			syscall[Offset + 6] == 0x40) //40h
		{
			//			__debugbreak();
			Handler = FixOpcodeTestInDriver;
		}

		// Fix lea r10,[nt!KeServiceDescriptorTable]
		// and lea r11,[nt!KeServiceDescriptorTableShadow]
		else if (syscall[Offset] == 0x4c &&
			syscall[Offset + 1] == 0x8d &&
			syscall[Offset + 7] == 0x4c &&
			syscall[Offset + 8] == 0x8d)
		{
			// Save their rel32
			leaR10Offset = Offset;
			leaR11Offset = Offset + 7;
		}

		// Fix lea rdi,[nt!KeServiceDescriptorTableShadow]
		else if (syscall[Offset] == 0x48 &&
			syscall[Offset + 1] == 0x8d &&
			syscall[Offset + 2] == 0x3d)
		{
			leaRdiOffset = Offset;
		}

		// Find call r10
		else if (syscall[Offset] == 0x41 &&
			syscall[Offset + 1] == 0xff &&
			syscall[Offset + 2] == 0xd2 &&
			syscall[Offset + 3] == 0x65 &&
			syscall[Offset + 4] == 0xff)
		{
			DbgPrint("Call r10 found!\n");
			CallR10InDriver = (ULONG_PTR)_syscall64 + Offset;
			CallR10InMdl = MdlVa + Offset;
			CallR10InSystem = SyscallOrigin + Offset;
			CallR10 = &CallR10InSystem;
		}

		if (Handler) {
			Handler((ULONG_PTR)_syscall64,
				SyscallOrigin,
				(ULONG_PTR)MdlVa,
				Offset,
				&Ptr);
			Handler = NULL;
		}
	}

	//	__debugbreak();

	if (!leaR10Offset || !leaR11Offset || !leaRdiOffset)
	{
		DbgPrint("Failed to initialize ssdt table!\n");
		return STATUS_NOT_FOUND;
	}
	else
	{
		RebuildSystemServiceTable((ULONG_PTR)_syscall64,
			SyscallOrigin,
			(ULONG_PTR)MdlVa,
			leaR10Offset,
			leaR11Offset,
			leaRdiOffset);
	}
	return status;
}

NTSTATUS RebuiltSsdt(ULONG_PTR SysSsdtAddress)
{
	PSERVICE_DESCRIPTOR_TABLE SysSsdt = (PSERVICE_DESCRIPTOR_TABLE)SysSsdtAddress;

	if (!SysSsdt)
		return STATUS_NOT_FOUND;

	// Fill SSDT in driver
	SsdtInDriver.ServiceTable = (PLONG)&SsdtTable;
	SsdtInDriver.ArgumentTable = SysSsdt->ArgumentTable;
	SsdtInDriver.CounterTable = SysSsdt->CounterTable;
	SsdtInDriver.TableSize = SysSsdt->TableSize;

	LONG TableRel;

	// Fill ssdt rel32
	for (USHORT index = 0;index < SysSsdt->TableSize;index++)
	{
		// address = ssdt[index] >> 4 + table_address
		TableRel = (LONG)SysSsdt->ServiceTable[index];
		SsdtAddress[index] = (TableRel >> 4) + (ULONG_PTR)SysSsdt->ServiceTable;

		if (SsdtAddress[index] == (ULONG_PTR)NtSetInformationThread ||
			SsdtAddress[index] == (ULONG_PTR)NtQueryInformationThread)
		{
			DbgPrint("Some functions found. Address = %p\n", SsdtAddress[index]);
			__debugbreak();

			// mov r10,qword ptr
			*(PULONG_PTR)(SsdtTrampolineInMdl) = 0x158B4C;

			// rel32
			ULONG32 Rel32 = (ULONG32)((ULONG_PTR)&SsdtAddress[index] - (SsdtTrampolineInDriver + 7));
			*(PULONG32)(SsdtTrampolineInMdl + 3) = Rel32;

			// add rsp, 8
			*(PULONG_PTR)(SsdtTrampolineInMdl + 7) = 0x08C48348;

			// jmp qword ptr nt!KiSystemServiceCopyEnd+0x10
			*(PULONG_PTR)(SsdtTrampolineInMdl + 11) = 0x25FF;

			// rel32
			Rel32 = (ULONG32)((ULONG_PTR)CallR10 - (SsdtTrampolineInDriver + 17));
			*(PULONG32)(SsdtTrampolineInMdl + 13) = Rel32;

			TableRel = ((LONG)(SsdtTrampolineInDriver - (ULONG_PTR)&SsdtTable)) << 4;
			SsdtTable[index] = TableRel;

			SsdtTrampolineInMdl += 17;
			SsdtTrampolineInDriver += 17;
		}
		else
		{
			// jmp qword ptr
			*(PULONG_PTR)SsdtTrampolineInMdl = 0x25FF;

			// rel32
			ULONG32 Rel32 = (ULONG32)((ULONG_PTR)&SsdtAddress[index] - (SsdtTrampolineInDriver + 6));
			*(PULONG32)(SsdtTrampolineInMdl + 2) = Rel32;

			TableRel = ((LONG)(SsdtTrampolineInDriver - (ULONG_PTR)&SsdtTable)) << 4;
			SsdtTable[index] = TableRel;

			SsdtTrampolineInMdl += 6;
			SsdtTrampolineInDriver += 6;
		}

	}
	return STATUS_SUCCESS;
}

NTSTATUS RebuiltShadowSsdt(ULONG_PTR SysSsdtAddress,
	ULONG_PTR SysShadowSsdtAddress)
{
	PSERVICE_DESCRIPTOR_TABLE SysSsdt = (PSERVICE_DESCRIPTOR_TABLE)SysSsdtAddress;
	PSERVICE_DESCRIPTOR_TABLE_SHADOW SysShadowSsdt = (PSERVICE_DESCRIPTOR_TABLE_SHADOW)SysShadowSsdtAddress;

	if (!SysSsdt || !SysShadowSsdt)
		return STATUS_NOT_FOUND;

	//	__debugbreak();

	PEPROCESS eprocess;
	LookupAWin32Process(&eprocess);
	if (!eprocess)
	{
		return STATUS_NOT_FOUND;
	}

	// Attach to a GUI process to obtain shadow ssdt address
	KAPC_STATE ApcState;
	KeStackAttachProcess(eprocess, &ApcState);

	// SSDT in Shadow SSDT
	ShadowSsdtInDriver.SsdtServiceTable = (PLONG)&SsdtTable;
	ShadowSsdtInDriver.SsdtArgumentTable = SysSsdt->ArgumentTable;
	ShadowSsdtInDriver.SsdtCounterTable = SysSsdt->CounterTable;
	ShadowSsdtInDriver.SsdtTableSize = SysSsdt->TableSize;

	// Shadow SSDT
	ShadowSsdtInDriver.ServiceTable = (PLONG)&ShadowSsdtTable;
	ShadowSsdtInDriver.ArgumentTable = SysShadowSsdt->ArgumentTable;
	ShadowSsdtInDriver.CounterTable = SysShadowSsdt->CounterTable;
	ShadowSsdtInDriver.TableSize = SysShadowSsdt->TableSize;

	LONG TableRel;

	// Fill shadow ssdt rel32
	for (USHORT index = 0;index < SysShadowSsdt->TableSize;index++)
	{
		// address = ssdt[index] >> 4 + table_address
		TableRel = (LONG)SysShadowSsdt->ServiceTable[index];

		// jmp qword ptr
		*(PULONG_PTR)ShadowSsdtTrampolineInMdl = 0x25FF;
		ShadowSsdtAddress[index] = (TableRel >> 4) + (ULONG_PTR)SysShadowSsdt->ServiceTable;

		// rel32
		ULONG32 Rel32 = (ULONG32)((ULONG_PTR)&ShadowSsdtAddress[index] - (ShadowSsdtTrampolineInDriver + 6));
		*(PULONG32)(ShadowSsdtTrampolineInMdl + 2) = Rel32;

		// Fill our new service table.
		TableRel = ((LONG)(ShadowSsdtTrampolineInDriver - (ULONG_PTR)&ShadowSsdtTable)) << 4;

		ShadowSsdtTable[index] = TableRel;

		ShadowSsdtTrampolineInDriver += 6;
		ShadowSsdtTrampolineInMdl += 6;
	}

	KeUnstackDetachProcess(&ApcState);

	return STATUS_SUCCESS;
}

VOID LookupAWin32Process(PEPROCESS *eprocess)
{
	PEPROCESS EProcess;
	for (USHORT pid = 4;;pid += 4)
	{
		if (PsLookupProcessByProcessId((HANDLE)pid, &EProcess) == STATUS_SUCCESS)
		{
			if (PsGetProcessWin32Process(EProcess))
			{
				*eprocess = EProcess;
				break;
			}
		}
	}
}

// Many bugs
NTSTATUS RebuildSystemServiceTable(ULONG_PTR DriverAddress,
	ULONG_PTR SystemAddress,
	ULONG_PTR MdlAddress,
	ULONG_PTR leaR10Offset,
	ULONG_PTR leaR11Offset,
	ULONG_PTR leaRdiOffset)
{
	if (!leaR10Offset || !leaR11Offset || !leaRdiOffset || !DriverAddress || !SystemAddress || !MdlAddress)
		return STATUS_NOT_FOUND;

	ULONG32 Rel32;

	// Obtain SSDT address
	Rel32 = *(PULONG32)(MdlAddress + leaR10Offset + 3);
	ULONG_PTR SysSsdt = SystemAddress + leaR10Offset + 7 + Rel32;
	if (Rel32)
	{
		RebuiltSsdt(SysSsdt);
		Rel32 = (ULONG32)((ULONG_PTR)&SsdtInDriver - (DriverAddress + leaR11Offset));
		*(PULONG32)(MdlAddress + leaR10Offset + 3) = Rel32;
	}
	else
	{
		return STATUS_NOT_FOUND;
	}

	// and Shadow SSDT
	Rel32 = *(PULONG32)(MdlAddress + leaR11Offset + 3);
	ULONG_PTR SysShadowSsdt = SystemAddress + leaR11Offset + 7 + Rel32;
	if (Rel32)
	{
		RebuiltShadowSsdt(SysSsdt, SysShadowSsdt);
		Rel32 = (ULONG32)((ULONG_PTR)&ShadowSsdtInDriver - (DriverAddress + leaR11Offset + 7));
		*(PULONG32)(MdlAddress + leaR11Offset + 3) = Rel32;
	}
	else
	{
		return STATUS_NOT_FOUND;
	}

	// Fix lea rdi, 
	Rel32 = *(PULONG32)(MdlAddress + leaRdiOffset + 3);
	if (Rel32)
	{
		Rel32 = (ULONG32)((ULONG_PTR)&ShadowSsdtInDriver - (DriverAddress + leaRdiOffset + 7));
		*(PULONG32)(MdlAddress + leaRdiOffset + 3) = Rel32;
	}
	else
	{
		return STATUS_NOT_FOUND;
	}

	return STATUS_SUCCESS;
}

VOID DriverUnload(PDRIVER_OBJECT DriverObject)
{
	UNREFERENCED_PARAMETER(DriverObject);

	if (Syscall64Pointer)
		ExFreePool(Syscall64Pointer);

	IoDeleteDevice(DriverObject->DeviceObject);
}

NTSTATUS DriverEntry(PDRIVER_OBJECT DriverObject, PUNICODE_STRING RegistryPath)
{
	UNREFERENCED_PARAMETER(RegistryPath);
	UNREFERENCED_PARAMETER(DriverObject);

	//
	// Set up major function pointers
	//
	DriverObject->DriverUnload = (PDRIVER_UNLOAD)DriverUnload;

	InitializeFakeSyscall();

	return STATUS_SUCCESS;
}


